# $Id: Request.pm 32060 2010-08-04 14:28:48Z wsl $
# $URL: https://infix.uvt.nl/its-id/trunk/sources/loandesk/lib/UvT/Loandesk/Request.pm $

use utf8;
use strict;
use warnings FATAL => 'all';

package UvT::Loandesk::Request;

use Xyzzy::Generator -Base;
use UvT::Loandesk::LDAP -mixin;

use UvT::Loandesk::Crypto;
use Time::HiRes qw(time);
use Aselect::Client;
use Aselect::Util;

field 'done';

# algemene idee:
# 1) parameters worden on-the-fly uitgerekend
# 2) elke mogelijke pagina bepaalt zelf of-ie genoeg info heeft
# 3) elke pagina breekt af indien done

sub getxml {
	my $url = shift;

	our $parser;
	unless(defined($parser)) {
		$parser = new XML::LibXML;
		$parser->validation(0);
		$parser->load_ext_dtd(0);
		$parser->expand_entities(0);
		$parser->pedantic_parser(1);
#		$parser->no_network(1);
	}
#	return $parser->parse_file($url);
	#my $start = time;
	my $xml = $parser->parse_file($url);
	#my $elapsed = time - $start;
	#warn sprintf("\%s (\%.2fs)\n", $url, $elapsed);
	return $xml;
}

sub error {
	my ($type, $msg) = @_;

	return if $self->done;
	$self->done(1);

	$self->init;
	$self->setup;

	$self->addTemplates('error');

	my $doc = $self->doc;

	my $error_type = $doc->getElementById('error_type');
	$error_type->appendText($type);

	my $error_msg = $doc->getElementById('error_msg');
	$error_msg->appendText($msg);

	$self->finalize;

	return;
}

sub form {
	my $cfg = $self->cfg;
	my $pica3 = $cfg->mps;

	my $epn = $self->epn or return;
	my $volume = $self->volume;
	my $xml = $self->getxml("$pica3?action=titleInfo&epn=$epn&format=d");
	my $title = $xml->findvalue('/xrep/pica3Response/item/kmc[@value="4000"]');
	$title =~ s/@//;

	return $self->error('Error', 'Unknown publication')
		unless $title;

#	$xml = $self->getxml("$pica3?action=listVolumes&epn=$epn");
#	my @vols = $xml->findnodes('/xrep/pica3Response/item');

#	return $self->error('Error', 'No volumes available')
#		unless @vols;

	my $back = $self->back;

	return if $self->done;
	$self->done(1);

	$self->init;
	$self->setup;

	$self->addTemplates('request', $self->requesttype);

	my $doc = $self->doc;
	my $content = $doc->getElementById('pubtitle');
	$content->appendText($title);

	my $session = $self->session;
	my $secret = $cfg->secret;
	my $nonce = loandesk_create_token('n'.$secret.$session, '');

	my $hidden = $doc->getElementById('nonce');
	$hidden->setAttribute('value', $nonce);

	$hidden = $doc->getElementById('epn');
	$hidden->setAttribute('value', $epn);

	$hidden = $doc->getElementById('back');
	if($back) {
		$hidden->setAttribute('value', $self->back);
	} else {
		$hidden->parentNode->removeChild($hidden);
		my $button = $doc->getElementById('close');
		$button->setAttribute(type => 'button');
	}

#	if(@vols > 1) {
#		my $volInfo = $doc->getElementById('volInfo');
#		foreach my $vol (@vols) {
#			my $key = $vol->findvalue('volumeID');
#			my $val = $vol->findvalue('volInfo');
#			my ($frag, $ids) = $self->getFragment('request', 'volume');
#			$ids->{input}->setAttribute('value', $key);
#			$ids->{input}->setAttribute('checked', 'checked')
#				if defined $volume && $key eq $volume;
#			$ids->{label}->appendText($val);
#			$volInfo->appendChild($frag);
#		}
#	} else {
#		my $vol = $vols[0];
#		my $key = $vol->findvalue('volumeID');
#		$hidden = $doc->createElement('input');
#		$hidden->setAttribute('type', 'hidden');
#		$hidden->setAttribute('name', 'volume');
#		$hidden->setAttribute('value', $key);
#		my $listVolumes = $doc->getElementById('listVolumes');
#		$listVolumes->replaceNode($hidden);
#	}
	$hidden = $doc->createElement('input');
	$hidden->setAttribute('type', 'hidden');
	$hidden->setAttribute('name', 'volume');
	$hidden->setAttribute('value', $volume);
	my $listVolumes = $doc->getElementById('listVolumes');
	$listVolumes->replaceNode($hidden);

	my $organization = $self->organization;
	if($organization) {
		my $radio = $doc->getElementById("organization_$organization");
		$radio->setAttribute('checked', 'checked')
			if defined $radio;

		my $login = $doc->getElementById('login');
		$login->parentNode->removeChild($login)
			if $organization eq 'extusr' && $self->authbroker;
	}

	$self->finalize;
}

sub aselect {
	my $req = $self->req;
	my $cfg = $self->cfg;
	my $app = $cfg->app_id;

	my $ticket = $req->cookie('aselect');
	if(defined $ticket) {
		my ($uid, $org) = eval { aselect_verify_ticket($ticket) };
		if($@) {
			warn "aselect_verify_ticket: $@";
		} else {
			return $uid;
		};
	}

	my %parms = (
			epn => $self->epn,
			volume => $self->volume,
			nonce => $self->nonce,
			action => 'request',
			submit => 1
		);
	my $back = $self->back;
	$parms{back} = $back if $back;

	my $url = aselect_url($req->url(-path=>1), \%parms);

	my $cred = $req->param('aselect_credentials');
	if(defined $cred) {
		my $rid = $req->param('rid');
		die "no rid param\n" unless defined $rid;
		eval { $ticket = aselect_verify_credentials($rid, $cred) };
		if($@) {
			$self->error('System error', 'A-Select login failed');
		} else {
			$req->setcookie(aselect => $ticket);
			$req->setheader(Location => $url);
			$req->status(302);
		};
		return undef;
	}

	return if $self->done;
	$self->done(1);

	my $sso = aselect_authenticate($app, $url);
	$req->setheader(Location => $sso);
	$req->status(302);

	return undef;
}

sub pica {
	my ($anr, $epn, $volume) = @_;
	my $cfg = $self->cfg;
	my $pica3 = $cfg->mps;

	my $pica3action = $self->pica3action;

	my $url = "$pica3?action=$pica3action&user=$anr&epn=$epn&volumeID=$volume";
	my $xml = $self->getxml($url);

	my $status = $xml->findvalue('/xrep/pica3Response/diagnostic/diagnosticCode');
	die "$url: Can't find response code in $pica3action reply\n"
		unless $status;
	my $message = $xml->findvalue('/xrep/pica3Response/diagnostic/diagnosticString');
	if($status eq 'OK') {
		return 1, $status, $message;
	} else {
		return undef, $status, $message
			if $message;
		my $krak = $xml->findvalue('/xrep/pica3Response/diagnostic/diagnosticAddInfo');
		if($krak) {
			die "$url: $status ($krak)\n";
		} else {
			die "$url: $status\n";
		}
	}
}

sub authbroker {
	return $self->{authbroker} if exists $self->{authbroker};
	my $req = $self->req;
	my $cfg = $self->cfg;
	my $secret = $cfg->secret;

	my $cookie = $req->cookie('authbroker');
	if($cookie) {
		my ($time, undef, $uid) = eval { loandesk_check_token('e'.$secret, $cookie) };
		return $self->{authbroker} = $uid
			if $uid && $time + $cfg->authbroker_expiry > time;
	}

	undef $self->{authbroker};

	my $uid = $req->param('userid')
		or return;
	my $pwd = $req->param('password')
		or return;

	my $sock = new IO::Socket::INET(
		PeerAddr => $cfg->authbroker_host,
		PeerPort => $cfg->authbroker_port,
		Timeout => $cfg->authbroker_timeout
	);

	die "can't connect to authbroker: $!\n"
		unless defined $sock;

	binmode($sock)
		or die "can't set socket binmode: $!\n";

	my $login = new XML::LibXML::Document('1.0', 'UTF-8');
	my $xrep = $login->createElement('xrep');
	$login->setDocumentElement($xrep);
	my $logonRequest = $login->createElement('logonRequest');
	$xrep->appendChild($logonRequest);

	$logonRequest->appendTextChild('userID', $uid);
	$logonRequest->appendTextChild('password', $pwd);
	$logonRequest->appendTextChild('requestSource', 'user');

	$login->toFH($sock, 1);

	my $parser = new XML::LibXML;
	$parser->pedantic_parser(1);
	$parser->validation(0);
	$parser->load_ext_dtd(0);
	$parser->expand_entities(0);
	$parser->keep_blanks(0);
	$parser->line_numbers(1);

	my $doc = $parser->parse_fh($sock);

	return unless $doc->find('/xrep/logonResponse/group="extusr"');

	my $session = loandesk_create_token('e'.$secret, $uid);
	$req->setcookie('authbroker', $session);

	return $self->{authbroker} = $uid;
}

sub anr {
	return $self->{anr} if exists $self->{anr};
	undef $self->{anr};
	my $org = $self->organization or return;
	$org = lc($org);

	my $anr;
	if($org eq 'uvt') {
		my $uid = $self->aselect or return;
		$anr = $self->ldap_anr($uid);
	} elsif($org eq 'extusr')  {
		$anr = $self->authbroker or return;
	} else {
		$self->error('Application error', "unknown organization '$org'");
	}
	return $self->{anr} = $anr;
}

sub request {
	my $cfg = $self->cfg;
	my $pica3 = $cfg->mps;

	return unless $self->submit;
	my $epn = $self->epn or return;
	my $volume = $self->volume or return;
	return unless $self->nonce;
	my $back = $self->back;
	return if $self->done;
	my $anr = $self->anr or return;
	return if $self->done;

	my ($ok, $status, $message) = $self->pica($anr, $epn, $volume);
	my $loans = $self->getxml("$pica3?action=listLoans&user=$anr");
	my $reservations = $self->getxml("$pica3?action=listReservations&user=$anr");

	$self->done(1);

	$self->init;
	$self->setup;
	$self->addTemplates('list');

	my $doc = $self->doc;

	my $feedback = $doc->getElementById('feedback');
	my ($frag, undef) = $self->getFragment($self->requesttype, $ok ? 'ok' : 'error');
	$frag->appendText($message);
	$feedback->replaceNode($frag);

	my @loantitles = $loans->findnodes('/xrep/pica3Response/item[loanStatus="aangevraagd"]/title');
	if(@loantitles) {
		my $list = $doc->getElementById('loans');
		foreach my $node (@loantitles) {
			my $title = $node->textContent;
			$title .= "…" if length($title) > 32;
			$list->appendTextChild('li', $title);
		}
	} else {
		my $list = $doc->getElementById('listLoans');
		$list->parentNode->removeChild($list);
	}

	my @reservationtitles = $reservations->findnodes('/xrep/pica3Response/item/title');
	if(@reservationtitles) {
		my $list = $doc->getElementById('reservations');
		foreach my $node (@reservationtitles) {
			my $title = $node->textContent;
			$title .= "…" if length($title) > 32;
			$list->appendTextChild('li', $title);
		}
	} else {
		my $list = $doc->getElementById('listReservations');
		$list->parentNode->removeChild($list);
	}

	if($back) {
		my $form = $doc->getElementById('back');
		$form->setAttribute(action => $back);
	} else {
		my $button = $doc->getElementById('close');
		$button->setAttribute(type => 'button');
	}

	$self->finalize;
}

sub back {
	return $self->{back} if exists $self->{back};
	my $req = $self->req;
	return $self->{back} = $req->param('back');
}

sub params {
	my $req = $self->req;
	foreach my $p ($req->param) {
		my $v = $req->param($p);
		warn "$p = $v\n";
	}
}

sub submit {
	return $self->{submit} if exists $self->{submit};
	my $req = $self->req;
	my $submit = $req->param('submit');
	return $self->{submit} = defined $submit;
}

sub action {
	return $self->{action} if exists $self->{action};
	my $req = $self->req;
	my $action = $req->param('action');
	if(defined $action && $action !~ /^\w+$/) {
		$self->error('Application error', 'action parameter not valid');
		undef $action;
	}
	return $self->{action} = $action;
}

sub epn {
	return $self->{epn} if exists $self->{epn};
	my $req = $self->req;
	my $epn = $req->param('epn');
	if(defined $epn) {
		unless($epn =~ /^\w+$/) {
			$self->error('Application error', 'epn parameter not valid');
			undef $epn;
		}
	} else {
		$self->error('Application error', 'epn parameter missing');
	}
	return $self->{epn} = $epn;
}

sub volume {
	return $self->{volume} if exists $self->{volume};
	my $req = $self->req;
	my $volume = $req->param('volume');
	if(defined $volume) {
		if($volume !~ /^\w+$/) {
			$self->error('Application error', 'volume parameter not valid');
			undef $volume;
		}
	} else {
		$self->error('Application error', 'volume parameter missing');
	}

#	if(defined $volume && $volume !~ /^\w+$/) {
#		$self->error('Application error', 'volume parameter not valid');
#		undef $volume;
#	}
	return $self->{volume} = $volume;
}

sub organization {
	return $self->{organization} if exists $self->{organization};
	my $req = $self->req;
	my $organization = $req->param('organization');
	if(defined $organization) {
		if($organization =~ /^\w+$/) {
			$req->setcookie('organization', $organization);
		} else {
			$self->error('Application error', 'organization parameter not valid');
		}
	} else {
		$organization = $req->cookie('organization');
		undef $organization
			if defined $organization && $organization !~ /^\w+$/;
	}
	return $self->{organization} = $organization;
}

sub session {
	return $self->{session} if exists $self->{session};
	my $req = $self->req;
	my $cfg = $self->cfg;

	my $secret = $cfg->secret;

	my $salt;
	eval {
		my $session = $req->cookie('session');
		(undef, $salt) = loandesk_check_token('s'.$secret, $session);

		my $nonce = $req->param('nonce');
		my ($time) = loandesk_check_token('n'.$secret.$salt, $nonce);
		die "expired nonce\n" if $time + $cfg->nonce_expiry < time;
		$self->{nonce} = $nonce;
	};

	unless(defined $salt) {
		(undef, $salt, my $session) = loandesk_create_token('s'.$secret, '');
		$req->setcookie(session => $session);
	}
	return $self->{session} = $salt;
}

sub nonce {
	return $self->{nonce} if exists $self->{nonce};
	undef $self->{nonce};
	# figuring out and checking the nonce is a side-effect of session():
	$self->session;
	return $self->{nonce};
}

sub goback {
	my $back = $self->back or return;
	my $action = $self->action or return;
	return if $action ne 'cancel';
	return unless $self->nonce;

	return if $self->done;
	$self->done(1);

	my $req = $self->req;
	$req->setheader(Location => $back);
	$req->status(302);
}

sub dispatch {
	# try things until one of them sticks:
	$self->goback;
	$self->request;
	$self->form;
}

sub work {
	eval { $self->dispatch(@_) };
	if($@) {
		warn $@;
		warn "died while done\n"
			if $self->done;
		$self->error('System error', 'an internal error occurred');
		$self->req->status(500);
	}
}
